from __future__ import annotations

import logging
from pathlib import Path
from typing import Optional, Union, get_args, get_origin, get_type_hints

import fiona
from fiona.transform import transform_geom
from rich.logging import RichHandler

from .classes import (
    AptFeature,
    AptMetadata,
    Boundary,
    LinearFeature,
    Pavement,
    Runway,
    Sign,
    StartupLocation,
    Windsock,
)
from .geometry import _DEFAULT_BEZIER_RESOLUTION
from .iterators import BIterator

try:
    from xplane_airports import AptDat
except ImportError as e:
    raise ImportError(
        "Could not import xplane_airports. Install https://github.com/X-Plane/xplane_airports."
    ) from e


logging.basicConfig(
    level=logging.INFO,
    format="%(message)s",
    datefmt="[%X]",
    handlers=[RichHandler(show_path=False, omit_repeated_times=False)],
)
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


_BASE_CRS = "EPSG:4326"

SUPPORTED_DRIVERS = {
    "ESRI Shapefile": {
        "multilayer": False,
        "extensions": ["shp", "dbf", "shz", "shp.zip"],
        "default_extension": "shp",
    },
    "FlatGeobuf": {
        "multilayer": False,
        "extensions": ["fgb"],
        "default_extension": "fgb",
    },
    "GeoJSON": {
        "multilayer": False,
        "extensions": ["json", "geojson"],
        "default_extension": "geojson",
    },
    "GeoJSONSeq": {
        "multilayer": False,
        "extensions": ["geojsonl", "geojsons"],
        "default_extension": "geojsonl",
    },
    "GPKG": {
        "multilayer": True,
        "extensions": ["gpkg"],
        "default_extension": "gpkg",
    },
    "GML": {
        "multilayer": False,
        "extensions": ["gml", "xml"],
        "default_extension": "gml",
    },
    "OGR_GMT": {
        "multilayer": False,
        "extensions": ["gmt"],
        "default_extension": "gmt",
    },
    "SQLite": {
        "multilayer": True,
        "extensions": ["sqlite", "db"],
        "default_extension": "sqlite",
    },
}


def _detect_driver(path: Path):
    """
    Attempt to auto-detect driver based on the extension
    """
    extension = path.suffix.lower().removeprefix(".")

    if not extension:
        return None

    for driver, driver_info in SUPPORTED_DRIVERS.items():
        driver_extensions = driver_info["extensions"]

        if extension in driver_extensions:
            return driver

    return None


VALID_FEATURES = [
    "boundary",
    "runways",
    "startup_locations",
    "windsocks",
    "signs",
    "pavements",
    "linear_features",
]


class ParsedAirport:
    id: str
    metadata: AptMetadata
    boundary: Optional[Boundary]
    runways: list[Runway]
    startup_locations: list[StartupLocation]
    signs: list[Sign]
    windsocks: list[Windsock]
    linear_features: list[LinearFeature]
    pavements: list[Pavement]

    def __init__(
        self,
        airport: AptDat.Airport,
        bezier_resolution: int = _DEFAULT_BEZIER_RESOLUTION,
    ) -> None:
        """A parsed X-Plane airport.

        Args:
            airport (xplane_airports.AptDat.Airport): An X-Plane airport object
                as obtained from `xplane_airports` (https://github.com/X-Plane/xplane_airports).
            bezier_resolution (int): Number of points to use to plot Bezier curves.
                A higher number means more resolution but also larger file sizes on export.
                Default 16.
        """
        self._airport = airport
        self.id = None
        self.metadata = AptMetadata()
        self.boundary = None
        self.runways = []
        self.startup_locations = []
        self.signs = []
        self.windsocks = []
        self.linear_features = []
        self.pavements = []

        self._parse(bezier_resolution=bezier_resolution)

    def _parse(self, bezier_resolution: int) -> None:
        logger.info("Parsing airport.")
        row_iterator = BIterator(self._airport.text)

        for row in row_iterator:
            row_code = row.row_code

            if row_code == AptDat.RowCode.AIRPORT_HEADER:
                self.id = row.tokens[4]

            if row_code == AptDat.RowCode.METADATA:
                self.metadata.add_from_row(row)

            elif row_code == AptDat.RowCode.BOUNDARY:
                logger.debug("Parsing boundary row.")

                boundary = Boundary.from_row_iterator(
                    row, row_iterator, bezier_resolution
                )

                if boundary is not None:
                    self.boundary = boundary

            elif row_code == AptDat.RowCode.LAND_RUNWAY:
                logger.debug("Parsing runway row.")

                runway = Runway.from_line(row)

                if runway is not None:
                    self.runways.append(runway)

            elif row_code == AptDat.RowCode.START_LOCATION_NEW:
                logger.debug("Parsing startup location row.")

                startup_location = StartupLocation.from_line(row)

                if startup_location is not None:
                    self.startup_locations.append(startup_location)

                # TODO: process RowCode.START_LOCATION_EXT (startup location metadata)

            elif row_code == AptDat.RowCode.WINDSOCK:
                logger.debug("Parsing sign row.")

                windsock = Windsock.from_line(row)

                if windsock is not None:
                    self.windsocks.append(windsock)

            elif row_code == AptDat.RowCode.TAXI_SIGN:
                logger.debug("Parsing sign row.")

                sign = Sign.from_line(row)

                if sign is not None:
                    self.signs.append(sign)

            elif row_code == AptDat.RowCode.TAXIWAY:
                logger.debug("Parsing pavement row.")

                pavement = Pavement.from_row_iterator(
                    row, row_iterator, bezier_resolution
                )

                if pavement is not None:
                    self.pavements.append(pavement)

            elif row_code == AptDat.RowCode.FREE_CHAIN:
                logger.debug("Parsing linear feature row.")

                for line in LinearFeature.from_row_iterator(
                    row, row_iterator, bezier_resolution
                ):
                    if line is not None:
                        self.linear_features.append(line)

    def export(
        self,
        output_path: Union[str, Path],
        driver: Optional[str] = None,
        crs: str = "EPSG:4326",
        features: list[str] = VALID_FEATURES,
    ) -> None:
        """Export the parsed airport in a GIS-friendly format.

        Args:
            output_path (str, Path): File path for the output.
                Note that with some drivers (like GeoJSON) multiple files will be generated (one per feature).
            driver (str): OGR driver (file format) to use for the output.
                See `xplane_apt_convert.base.SUPPORTED_DRIVERS.keys()` for a list of valid options.
                If `None`, then the driver will be automatically inferred from the file extension in `output_path`.
                Default `None`.
            crs (str): Coordinate reference system (CRS) to use for the output.
                Default "EPSG:4326".
            features (list[str]): List of feature layers to include in the output.
                See `xplane_apt_convert.base.VALID_FEATURES` for a list of valid options.
                Default is all.
        """
        if driver is not None and driver not in SUPPORTED_DRIVERS:
            raise ValueError(
                f"Invalid driver '{driver}'. Supported drivers: {list(SUPPORTED_DRIVERS.keys())}"
            )

        if (
            isinstance(output_path, str)
            and len(output_path) > 0
            and output_path[-1] == "/"
        ):
            raise ValueError("output_path cannot be a directory.")

        if not isinstance(output_path, Path):
            output_path = Path(output_path)

        suffix = output_path.suffix

        if not suffix and not driver:
            raise ValueError(
                "Unknown output driver. Output path must have a valid file extension or an explicit driver must be provided."
            )

        if driver is None:
            driver = _detect_driver(output_path)

            if driver is None:
                raise ValueError(
                    f"Format {output_path.suffix} not recognized. Provide a valid file extension or specify an explicit driver."
                )

        base_folder = output_path.parent
        base_name = output_path.stem

        if not suffix:
            suffix = "." + SUPPORTED_DRIVERS[driver]["default_extension"]

        base_folder.mkdir(parents=True, exist_ok=True)

        for part_name in features:
            if part_name not in VALID_FEATURES:
                raise ValueError(
                    f"{part_name} is not a valid include. Valid options are: {VALID_FEATURES}"
                )

            part = getattr(self, part_name)

            if not part:
                logger.warning(f"No {part_name} found.")
                continue

            logger.info(f"Writing {part_name}.")

            part_type = get_type_hints(ParsedAirport)[part_name]
            part_type_origin = get_origin(part_type)
            part_type_args = get_args(part_type)

            part_is_list = get_origin(part_type) == list

            if part_type_origin == list:
                part_is_list = True
                schema = part_type_args[0]._schema()
            elif part_type_origin == Union:
                assert len(part_type_args) == 2 and part_type_args[1] == type(
                    None
                )  # Only Unions equivalent to Optional are valid.

                part_is_list = False
                schema = part_type_args[0]._schema()
            else:
                part_is_list = False
                schema = part_type._schema()

            multilayer = SUPPORTED_DRIVERS[driver]["multilayer"]

            if multilayer:
                path = base_folder / f"{base_name}{suffix}"
                layer = part_name
            else:
                path = base_folder / f"{base_name}.{part_name}{suffix}"
                layer = None

            with fiona.open(
                path,
                mode="w",
                crs=crs,
                driver=driver,
                schema=schema,
                layer=layer,
            ) as output:
                if part_is_list:
                    records = [record._to_record() for record in part]
                else:
                    records = [part._to_record()]

                if _BASE_CRS != crs:
                    for record in records:
                        record["geometry"] = transform_geom(
                            _BASE_CRS, crs, record["geometry"]
                        )

                output.writerecords(records)



from bezier.curve import Curve
from xplane_airports.AptDat import RowCode
import numpy as np


BEZIER_RESOLUTION = 15


def _calculate_quadratic_bezier(p0, p1, p2, resolution=BEZIER_RESOLUTION):
    if p0 == p2:
        return [p0]

    curve = Curve.from_nodes(np.asarray([p0, p1, p2]).T)
    curve_points = curve.evaluate_multi(np.linspace(0.0, 1.0, resolution))

    return curve_points.T.tolist()


def _calculate_cubic_bezier(p0, p1, p2, p3, resolution=BEZIER_RESOLUTION):
    if p0 == p3:
        return [p0]

    curve = Curve.from_nodes(np.asarray([p0, p1, p2, p3]).T)
    curve_points = curve.evaluate_multi(np.linspace(0.0, 1.0, resolution))

    return curve_points.T.tolist()


def _calculate_bezier(p0, p1, p2, p3=None, resolution=BEZIER_RESOLUTION):
    if p3 is None:
        return _calculate_quadratic_bezier(p0, p1, p2)
    else:
        return _calculate_cubic_bezier(p0, p1, p2, p3)


def get_path(row_iterator):
    def _process_row(is_bezier, tokens):
        nonlocal in_bezier, temp_bezier_nodes, coordinates, properties

        lat, lon = float(tokens[1]), float(tokens[2])

        if not is_bezier:
            if len(tokens) > 3:
                properties["painted_line_type"] = int(tokens[3])
            if len(tokens) > 4:
                properties["lighting_line_type"] = int(tokens[4])

            if in_bezier:
                temp_bezier_nodes.append((lon, lat))
                coordinates.extend(_calculate_bezier(*temp_bezier_nodes))
                temp_bezier_nodes = []
            else:
                coordinates.append((lon, lat))

            in_bezier = False

        else:
            bzp_lat, bzp_lon = float(tokens[3]), float(tokens[4])

            if len(tokens) > 5:
                properties["painted_line_type"] = int(tokens[5])
            if len(tokens) > 6:
                properties["lighting_line_type"] = int(tokens[6])

            if in_bezier:
                diff_lat = bzp_lat - lat
                diff_lon = bzp_lon - lon
                mirr_lat = lat - diff_lat
                mirr_lon = lon - diff_lon

                temp_bezier_nodes.append((mirr_lon, mirr_lat))
                temp_bezier_nodes.append((lon, lat))
                coordinates.extend(_calculate_bezier(*temp_bezier_nodes))
                temp_bezier_nodes = []
            else:
                if len(coordinates):
                    diff_lat = bzp_lat - lat
                    diff_lon = bzp_lon - lon
                    mirr_lat = lat - diff_lat
                    mirr_lon = lon - diff_lon

                    temp_bezier_nodes.append(coordinates[-1])
                    temp_bezier_nodes.append((mirr_lon, mirr_lat))
                    temp_bezier_nodes.append((lon, lat))
                    coordinates.extend(_calculate_bezier(*temp_bezier_nodes))
                    temp_bezier_nodes = []

            temp_bezier_nodes.append((lon, lat))
            temp_bezier_nodes.append((bzp_lon, bzp_lat))

            # else:
            in_bezier = True

    coordinates_list = []
    properties_list = []
    more_segments = True

    while more_segments:
        coordinates = []
        properties = {}
        temp_bezier_nodes = []
        in_bezier = False
        first_row = None
        first_row_is_bezier = None

        for row in row_iterator:
            if first_row is None:
                first_row = row
                first_row_is_bezier = row.row_code == RowCode.LINE_CURVE

            row_code = row.row_code
            tokens = row.tokens

            if row_code == RowCode.LINE_SEGMENT:
                _process_row(False, tokens)
            elif row_code == RowCode.LINE_CURVE:
                _process_row(True, tokens)
            elif row_code == RowCode.RING_SEGMENT:
                _process_row(False, tokens)
                _process_row(first_row_is_bezier, first_row.tokens)
                break
            elif row_code == RowCode.RING_CURVE:
                _process_row(True, tokens)
                _process_row(first_row_is_bezier, first_row.tokens)
                break
            elif row_code == RowCode.END_SEGMENT:
                _process_row(False, tokens)
                break
            elif row_code == RowCode.END_CURVE:
                _process_row(True, tokens)
                break
            else:
                # raise ValueError(f"Unexpected row code {row.row_code}")
                row_iterator.unnext()
                more_segments = False
                break
        else:
            # there is no more rows
            more_segments = False

        if len(coordinates) > 0:
            coordinates_list.append(coordinates)
            properties_list.append(properties)

    return coordinates_list, properties_list


def get_path_features(row_iterator):
    # https://forums.x-plane.org/index.php?/forums/topic/66713-understanding-the-logic-of-bezier-control-points-in-aptdat/

    # TODO: One same line can have different line styles (different properties), should be split into different features.
    # features = []

    coordinates, properties = get_path(row_iterator)

    assert len(coordinates) == len(properties)

    # props.update(
    #     {
    #         "type": "line",
    #     }
    # )

    # linear_feature = geojson.Feature(geometry=line, properties=props)
    # feature = {
    #     "geometry": {
    #         "type": "LineString",
    #         "coordinates": node_paths[0],
    #     },
    #     "properties": {},
    # }

    # features.append(feature)

    return coordinates, properties



"""

Demonstrates how to create a dynamic group of progress bars,
showing multi-level progress for multiple tasks (installing apps in the example),
each of which consisting of multiple steps.

"""

import time

from rich.console import Group
from rich.panel import Panel
from rich.live import Live
from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TextColumn,
    TimeElapsedColumn,
)


def run_steps(name, step_times, app_steps_task_id):
    """Run steps for a single app, and update corresponding progress bars."""

    for idx, step_time in enumerate(step_times):
        # add progress bar for this step (time elapsed + spinner)
        action = step_actions[idx]
        step_task_id = step_progress.add_task("", action=action, name=name)

        # run steps, update progress
        for _ in range(step_time):
            time.sleep(0.5)
            step_progress.update(step_task_id, advance=1)

        # stop and hide progress bar for this step when done
        step_progress.stop_task(step_task_id)
        step_progress.update(step_task_id, visible=False)

        # also update progress bar for current app when step is done
        app_steps_progress.update(app_steps_task_id, advance=1)


# progress bar for current app showing only elapsed time,
# which will stay visible when app is installed
current_app_progress = Progress(
    TimeElapsedColumn(),
    TextColumn("{task.description}"),
)

# progress bars for single app steps (will be hidden when step is done)
step_progress = Progress(
    TextColumn("  "),
    TimeElapsedColumn(),
    TextColumn("[bold purple]{task.fields[action]}"),
    SpinnerColumn("simpleDots"),
)
# progress bar for current app (progress in steps)
app_steps_progress = Progress(
    TextColumn(
        "[bold blue]Progress for app {task.fields[name]}: {task.percentage:.0f}%"
    ),
    BarColumn(),
    TextColumn("({task.completed} of {task.total} steps done)"),
)
# overall progress bar
overall_progress = Progress(
    TimeElapsedColumn(), BarColumn(), TextColumn("{task.description}")
)
# group of progress bars;
# some are always visible, others will disappear when progress is complete
progress_group = Group(
    Panel(Group(current_app_progress, step_progress, app_steps_progress)),
    overall_progress,
)

# tuple specifies how long each step takes for that app
step_actions = ("downloading", "configuring", "building", "installing")
apps = [
    ("one", (2, 1, 4, 2)),
    ("two", (1, 3, 8, 4)),
    ("three", (2, 1, 3, 2)),
]

# create overall progress bar
overall_task_id = overall_progress.add_task("", total=len(apps))

# use own live instance as context manager with group of progress bars,
# which allows for running multiple different progress bars in parallel,
# and dynamically showing/hiding them
with Live(progress_group):

    for idx, (name, step_times) in enumerate(apps):
        # update message on overall progress bar
        top_descr = "[bold #AAAAAA](%d out of %d apps installed)" % (idx, len(apps))
        overall_progress.update(overall_task_id, description=top_descr)

        # add progress bar for steps of this app, and run the steps
        current_task_id = current_app_progress.add_task("Installing app %s" % name)
        app_steps_task_id = app_steps_progress.add_task(
            "", total=len(step_times), name=name
        )
        run_steps(name, step_times, app_steps_task_id)

        # stop and hide steps progress bar for this specific app
        app_steps_progress.update(app_steps_task_id, visible=False)
        current_app_progress.stop_task(current_task_id)
        current_app_progress.update(
            current_task_id, description="[bold green]App %s installed!" % name
        )

        # increase overall progress now this task is done
        overall_progress.update(overall_task_id, advance=1)

    # final update for message on overall progress bar
    overall_progress.update(
        overall_task_id, description="[bold green]%s apps installed, done!" % len(apps)
    )



import numpy as np
from bezier.curve import Curve
from xplane_airports.AptDat import RowCode

_DEFAULT_BEZIER_RESOLUTION = 16


def _calculate_quadratic_bezier(p0, p1, p2, resolution):
    if p0 == p2:
        return [p0]

    curve = Curve.from_nodes(np.asarray([p0, p1, p2]).T)
    curve_points = curve.evaluate_multi(np.linspace(0.0, 1.0, resolution))

    return curve_points.T.tolist()


def _calculate_cubic_bezier(p0, p1, p2, p3, resolution):
    if p0 == p3:
        return [p0]

    curve = Curve.from_nodes(np.asarray([p0, p1, p2, p3]).T)
    curve_points = curve.evaluate_multi(np.linspace(0.0, 1.0, resolution))

    return curve_points.T.tolist()


def _calculate_bezier(p0, p1, p2, p3=None, resolution=_DEFAULT_BEZIER_RESOLUTION):
    if p3 is None:
        return _calculate_quadratic_bezier(p0, p1, p2, resolution)
    else:
        return _calculate_cubic_bezier(p0, p1, p2, p3, resolution)


def get_paths(row_iterator, bezier_resolution, mode="line"):
    # https://forums.x-plane.org/index.php?/forums/topic/66713-understanding-the-logic-of-bezier-control-points-in-aptdat/

    assert mode == "line" or mode == "polygon"

    coordinates = []
    properties = {}

    def _start_segment():
        nonlocal coordinates, properties
        coordinates = []
        properties = {}

    def _finish_segment():
        nonlocal coordinates, properties
        if len(coordinates) > 1:
            # simplify line. remove consecutive duplicates
            prev_c = None
            fixed_coordinates = []
            for c in coordinates:
                if prev_c is not None and tuple(c) == tuple(prev_c):
                    continue

                fixed_coordinates.append(c)

                prev_c = c

            coordinates_list.append(fixed_coordinates)
            properties_list.append(properties)

    def _process_row(is_bezier, tokens):
        nonlocal in_bezier, temp_bezier_nodes, coordinates, properties
        lat, lon = float(tokens[1]), float(tokens[2])

        if not is_bezier:
            if in_bezier:
                temp_bezier_nodes.append((lon, lat))
                coordinates.extend(
                    _calculate_bezier(*temp_bezier_nodes)
                )  # TODO: pass resolution argument
                temp_bezier_nodes = []
            else:
                coordinates.append((lon, lat))

            in_bezier = False

            painted_line_type = int(tokens[3]) if len(tokens) > 3 else None
            lighting_line_type = int(tokens[4]) if len(tokens) > 4 else None

            if mode == "line" and (
                (
                    painted_line_type is not None
                    and properties.get("painted_line_type") is not None
                    and painted_line_type != properties["painted_line_type"]
                )
                or (
                    lighting_line_type is not None
                    and properties.get("lighting_line_type") is not None
                    and lighting_line_type != properties["lighting_line_type"]
                )
            ):
                if row_iterator.has_next():
                    _finish_segment()
                    _start_segment()
                    row_iterator.unnext()  # reuse row for the new segment
            else:
                if painted_line_type is not None:
                    properties["painted_line_type"] = painted_line_type

                if lighting_line_type is not None:
                    properties["lighting_line_type"] = lighting_line_type

        else:
            bzp_lat, bzp_lon = float(tokens[3]), float(tokens[4])

            if in_bezier:
                diff_lat = bzp_lat - lat
                diff_lon = bzp_lon - lon
                mirr_lat = lat - diff_lat
                mirr_lon = lon - diff_lon

                temp_bezier_nodes.append((mirr_lon, mirr_lat))
                temp_bezier_nodes.append((lon, lat))
                coordinates.extend(
                    _calculate_bezier(*temp_bezier_nodes, resolution=bezier_resolution)
                )
                temp_bezier_nodes = []
            else:
                if len(coordinates):
                    diff_lat = bzp_lat - lat
                    diff_lon = bzp_lon - lon
                    mirr_lat = lat - diff_lat
                    mirr_lon = lon - diff_lon

                    temp_bezier_nodes.append(coordinates[-1])
                    temp_bezier_nodes.append((mirr_lon, mirr_lat))
                    temp_bezier_nodes.append((lon, lat))
                    coordinates.extend(
                        _calculate_bezier(
                            *temp_bezier_nodes, resolution=bezier_resolution
                        )
                    )
                    temp_bezier_nodes = []

            temp_bezier_nodes.append((lon, lat))
            temp_bezier_nodes.append((bzp_lon, bzp_lat))

            # else:
            in_bezier = True

            painted_line_type = int(tokens[5]) if len(tokens) > 5 else None
            lighting_line_type = int(tokens[6]) if len(tokens) > 6 else None

            if mode == "line" and (
                (
                    painted_line_type is not None
                    and properties.get("painted_line_type") is not None
                    and painted_line_type != properties["painted_line_type"]
                )
                or (
                    lighting_line_type is not None
                    and properties.get("lighting_line_type") is not None
                    and lighting_line_type != properties["lighting_line_type"]
                )
            ):
                if row_iterator.has_next():
                    _finish_segment()
                    _start_segment()
                    row_iterator.unnext()  # reuse row for the new segment
            else:
                if painted_line_type is not None:
                    properties["painted_line_type"] = painted_line_type

                if lighting_line_type is not None:
                    properties["lighting_line_type"] = lighting_line_type

    coordinates_list = []
    properties_list = []
    more_segments = True

    while more_segments:
        temp_bezier_nodes = []
        in_bezier = False
        first_row = None
        first_row_is_bezier = None

        _start_segment()

        for row in row_iterator:
            if first_row is None:
                first_row = row
                first_row_is_bezier = row.row_code in [
                    RowCode.LINE_CURVE,
                    RowCode.RING_CURVE,
                    RowCode.END_CURVE,
                ]

            row_code = row.row_code
            tokens = row.tokens

            if row_code == RowCode.LINE_SEGMENT:
                _process_row(False, tokens)
            elif row_code == RowCode.LINE_CURVE:
                _process_row(True, tokens)
            elif row_code == RowCode.RING_SEGMENT:
                _process_row(False, tokens)
                _process_row(first_row_is_bezier, first_row.tokens)
                break
            elif row_code == RowCode.RING_CURVE:
                _process_row(True, tokens)
                _process_row(first_row_is_bezier, first_row.tokens)
                break
            elif row_code == RowCode.END_SEGMENT:
                _process_row(False, tokens)
                break
            elif row_code == RowCode.END_CURVE:
                _process_row(True, tokens)
                break
            else:
                row_iterator.unnext()
                more_segments = False
                break
        else:
            # there is no more rows
            more_segments = False

        _finish_segment()

    assert len(coordinates_list) == len(properties_list)
    return coordinates_list, properties_list



class BIterator:
    def __init__(self, seq):
        self._seq = seq
        self._idx = -1

    def __iter__(self):
        return self

    def __next__(self):
        self._idx += 1

        if self._idx >= len(self._seq):
            self._idx -= 1
            raise StopIteration()

        return self._seq[self._idx]

    def unnext(self):
        self._idx -= 1

        if self._idx < -1:
            self._idx += 1
            raise IndexError("Can't go back any further.")

    def has_next(self):
        return self._idx < (len(self._seq) - 1)



from abc import ABC, abstractmethod
from collections import OrderedDict
from dataclasses import dataclass
from enum import Enum

from .geometry import get_paths
from .iterators import BIterator

try:
    from xplane_airports import AptDat
except ImportError as e:
    raise ImportError(
        "Could not import xplane_airports. See https://github.com/X-Plane/xplane_airports."
    ) from e


class SurfaceType(Enum):
    ASPHALT = 1  # Asphalt
    CONCRETE = 2  # Concrete
    TURF_OR_GRASS = 3  # Turf or grass
    DIRT = 4  # Dirt (brown)
    GRAVEL = 5  # Gravel (grey)
    DRY_LAKEBED = 12  # Dry lakebed
    WATER_RUNWAY = 13  # Water runway
    SNOW_OR_ICE = 14  # Snow or ice
    TRANSPARENT = 15  # Transparent


class ShoulderSurfaceType(Enum):
    NONE = 0  # No shoulder
    ASPHALT = 1  # Asphalt shoulder
    CONCRETE = 2  # Concrete shoulder


class RunwayMarking(Enum):
    NONE = 0  # No runway markings
    VISUAL = 1  # Visual markings
    NON_PRECISION = 3  # Non-precision approach markings
    PRECISION = 4  # Precision approach markings
    UK_NON_PRECISION = 5  # UK-style non-precision approach markings
    UK_PRECISION = 5  # UK-style precision approach markings


class SignSize(Enum):
    SMALL = 1  # Small taxiway sign
    MEDIUM = 2  # Medium taxiway sign
    LARGE = 3  # Large taxiway sign
    LARGE_DISTANCE_REMAINING = 4  # Large distance-remaining sign on runway edge
    SMALL_DISTANCE_REMAINING = 5  # Small distance-remaining sign on runway edg


class AptMetadata(dict):
    def add_from_row(self, row: AptDat.AptDatLine):
        tokens = row.tokens

        key = tokens[1]
        value = tokens[2] if len(tokens) > 2 else None

        self[key] = value


class AptFeature(ABC):
    @staticmethod
    @abstractmethod
    def _schema() -> dict:
        pass

    @abstractmethod
    def _to_record(self) -> dict:
        pass


@dataclass
class Boundary(AptFeature):
    name: str
    coordinates: list[tuple[float, float]]

    @staticmethod
    def from_row_iterator(
        header_row: AptDat.AptDatLine, line_iterator: BIterator, bezier_resolution: int
    ) -> "Boundary":
        tokens = header_row.tokens
        coordinates_list, properties_list = get_paths(
            line_iterator,
            bezier_resolution=bezier_resolution,
            mode="polygon",
        )

        coordinates_list = [c for c in coordinates_list if len(c) > 2]
        properties_list = [
            p for c, p in zip(coordinates_list, properties_list) if len(c) > 2
        ]

        return Boundary(
            name=" ".join(tokens[1:]),
            coordinates=coordinates_list,
        )

    @staticmethod
    def _schema():
        return {
            "geometry": "Polygon",
            "properties": OrderedDict(
                [
                    ("name", "str"),
                ]
            ),
        }

    def _to_record(self):
        return {
            "geometry": {
                "type": "Polygon",
                "coordinates": self.coordinates,
            },
            "properties": {"name": self.name},
        }


@dataclass
class Pavement(AptFeature):
    surface_type: SurfaceType
    smoothness: float
    texture_orientation: float
    name: str
    coordinates: list[tuple[float, float]]

    @staticmethod
    def from_row_iterator(
        header_row: AptDat.AptDatLine, line_iterator: BIterator, bezier_resolution: int
    ) -> "Pavement":
        tokens = header_row.tokens
        coordinates_list, properties_list = get_paths(
            line_iterator,
            bezier_resolution=bezier_resolution,
            mode="polygon",
        )

        coordinates_list = [c for c in coordinates_list if len(c) > 2]
        properties_list = [
            p for c, p in zip(coordinates_list, properties_list) if len(c) > 2
        ]

        return Pavement(
            surface_type=SurfaceType(int(tokens[1])),
            smoothness=float(tokens[2]),
            texture_orientation=float(tokens[3]),
            name=" ".join(tokens[4:]),
            coordinates=coordinates_list,
        )

    @staticmethod
    def _schema():
        return {
            "geometry": "Polygon",
            "properties": OrderedDict(
                [
                    ("name", "str"),
                    ("surface_type", "str"),
                    ("smoothness", "float"),
                    ("texture_orientation", "float"),
                    # TODO: pavements can have edge lines,
                    # but they can cover only a segment of the edge,
                    # and/or have multiple types in different parts,
                    # so this can't be a pavement property.
                    # Probably consider making separate line features.
                    # ("painted_line_type", "int"),
                    # ("lighting_line_type", "int"),
                ]
            ),
        }

    def _to_record(self):
        return {
            "geometry": {
                "type": "Polygon",
                "coordinates": self.coordinates,
            },
            "properties": {
                "name": self.name,
                "surface_type": self.surface_type.name,
                "smoothness": self.smoothness,
                "texture_orientation": self.texture_orientation,
            },
        }


@dataclass
class LinearFeature(AptFeature):
    name: str
    painted_line_type: int
    lighting_line_type: int
    coordinates: list[tuple[float, float]]

    @staticmethod
    def from_row_iterator(
        header_row: AptDat.AptDatLine,
        line_iterator: BIterator,
        bezier_resolution: int,
    ) -> list["LinearFeature"]:
        tokens = header_row.tokens
        coordinates_list, properties_list = get_paths(
            line_iterator,
            bezier_resolution=bezier_resolution,
            mode="line",
        )

        return [
            LinearFeature(
                name=" ".join(tokens[1:]),
                painted_line_type=properties.get("painted_line_type"),
                lighting_line_type=properties.get("lighting_line_type"),
                coordinates=coordinates,
            )
            for coordinates, properties in zip(coordinates_list, properties_list)
            if len(coordinates) > 1
        ]

    @staticmethod
    def _schema():
        return {
            "geometry": "LineString",
            "properties": OrderedDict(
                [
                    ("name", "str"),
                    ("painted_line_type", "int"),
                    ("lighting_line_type", "int"),
                ]
            ),
        }

    def _to_record(self):
        return {
            "geometry": {
                "type": "LineString",
                "coordinates": self.coordinates,
            },
            "properties": {
                "name": self.name,
                "painted_line_type": self.painted_line_type,
                "lighting_line_type": self.lighting_line_type,
            },
        }


@dataclass
class RunwayEnd:
    name: str
    latitude: float
    longitude: float
    dthr_length: float  # Displaced Threshold length in meters
    overrun_length: float  # in meters
    marking: RunwayMarking
    lighting: int
    tdz_lighting: int  # Touchdown Zone lighting
    reil: int  # Runway End Identifier Lights

    @staticmethod
    def from_line_tokens(tokens: list[str]) -> "RunwayEnd":
        return RunwayEnd(
            name=tokens[0],
            latitude=float(tokens[1]),
            longitude=float(tokens[2]),
            dthr_length=float(tokens[3]),
            overrun_length=float(tokens[4]),
            marking=None,  # TODO
            lighting=None,  # TODO
            tdz_lighting=None,  # TODO
            reil=None,  # TODO
        )


@dataclass
class Runway(AptFeature):
    width: float  # in meters
    surface_type: SurfaceType
    shoulder_surface_type: ShoulderSurfaceType
    smoothness: float
    centerline_lights: int
    edge_lights: int
    auto_distance_remaining_signs: bool
    ends: tuple[RunwayEnd, RunwayEnd]

    @staticmethod
    def from_line(line: AptDat.AptDatLine) -> "Runway":
        tokens = line.tokens
        return Runway(
            width=float(tokens[1]),
            surface_type=SurfaceType(int(tokens[2])),
            shoulder_surface_type=ShoulderSurfaceType(int(tokens[3])),
            smoothness=float(tokens[4]),
            centerline_lights=int(tokens[5]),
            edge_lights=int(tokens[6]),
            auto_distance_remaining_signs=bool(int(tokens[7])),
            ends=(
                RunwayEnd.from_line_tokens(tokens[8:17]),
                RunwayEnd.from_line_tokens(tokens[17:26]),
            ),
        )

    @staticmethod
    def _schema():
        return {
            "geometry": "LineString",
            "properties": OrderedDict(
                [
                    ("width", "float"),
                    ("surface_type", "str"),
                    ("shoulder_surface_type", "str"),
                    ("smoothness", "float"),
                    ("centerline_lights", "int"),
                    ("edge_lights", "int"),
                    ("auto_distance_remaining_signs", "bool"),
                ]
            ),
        }

    def _to_record(self):
        return {
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    (self.ends[0].longitude, self.ends[0].latitude),
                    (self.ends[1].longitude, self.ends[1].latitude),
                ],
            },
            "properties": {
                "width": self.width,
                "surface_type": self.surface_type.name,
                "shoulder_surface_type": self.shoulder_surface_type.name,
                "smoothness": self.smoothness,
                "centerline_lights": self.centerline_lights,
                "edge_lights": self.edge_lights,
                "auto_distance_remaining_signs": self.auto_distance_remaining_signs,
            },
        }


@dataclass
class StartupLocation(AptFeature):
    latitude: float
    longitude: float
    heading: float
    location_type: str
    airplane_types: str
    name: str

    @staticmethod
    def from_line(line: AptDat.AptDatLine) -> "StartupLocation":
        tokens = line.tokens
        return StartupLocation(
            latitude=float(tokens[1]),
            longitude=float(tokens[2]),
            heading=float(tokens[3]),
            location_type=tokens[4],
            airplane_types=tokens[5],
            name=" ".join(tokens[6:]),
        )

    @staticmethod
    def _schema():
        return {
            "geometry": "Point",
            "properties": OrderedDict(
                [
                    ("heading", "float"),
                    ("location_type", "str"),
                    ("airplane_types", "str"),
                    ("name", "str"),
                ]
            ),
        }

    def _to_record(self):
        return {
            "geometry": {
                "type": "Point",
                "coordinates": (self.longitude, self.latitude),
            },
            "properties": {
                "heading": self.heading,
                "location_type": self.location_type,
                "airplane_types": self.airplane_types,
                "name": self.name,
            },
        }


@dataclass
class Windsock(AptFeature):
    latitude: float
    longitude: float
    illuminated: bool
    name: str

    @staticmethod
    def from_line(line: AptDat.AptDatLine) -> "Windsock":
        tokens = line.tokens
        return Windsock(
            latitude=float(tokens[1]),
            longitude=float(tokens[2]),
            illuminated=bool(int(tokens[3])),
            name=" ".join(tokens[4:]),
        )

    @staticmethod
    def _schema():
        return {
            "geometry": "Point",
            "properties": OrderedDict(
                [
                    ("illuminated", "bool"),
                    ("name", "str"),
                ]
            ),
        }

    def _to_record(self):
        return {
            "geometry": {
                "type": "Point",
                "coordinates": (self.longitude, self.latitude),
            },
            "properties": {
                "illuminated": self.illuminated,
                "name": self.name,
            },
        }


@dataclass
class Sign(AptFeature):
    latitude: float
    longitude: float
    heading: float
    size: SignSize
    text: str

    @staticmethod
    def from_line(line: AptDat.AptDatLine) -> "Sign":
        tokens = line.tokens
        return Sign(
            latitude=float(tokens[1]),
            longitude=float(tokens[2]),
            heading=float(tokens[3]),
            size=SignSize(int(tokens[5])),
            text=tokens[6],
        )

    @staticmethod
    def _schema():
        return {
            "geometry": "Point",
            "properties": OrderedDict(
                [
                    ("heading", "float"),
                    ("size", "str"),
                    ("text", "str"),
                ]
            ),
        }

    def _to_record(self):
        return {
            "geometry": {
                "type": "Point",
                "coordinates": (self.longitude, self.latitude),
            },
            "properties": {
                "heading": self.heading,
                "size": self.size.name,
                "text": self.text,
            },
        }



"""

Demonstrates the use of multiple Progress instances in a single Live display.    

"""

from time import sleep

from rich.live import Live
from rich.panel import Panel
from rich.progress import Progress, SpinnerColumn, BarColumn, TextColumn
from rich.table import Table


job_progress = Progress(
    "{task.description}",
    SpinnerColumn(),
    BarColumn(),
    TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
)
job1 = job_progress.add_task("[green]Cooking")
job2 = job_progress.add_task("[magenta]Baking", total=200)
job3 = job_progress.add_task("[cyan]Mixing", total=400)

total = sum(task.total for task in job_progress.tasks)
overall_progress = Progress()
overall_task = overall_progress.add_task("All Jobs", total=int(total))

progress_table = Table.grid()
progress_table.add_row(
    Panel.fit(
        overall_progress, title="Overall Progress", border_style="green", padding=(2, 2)
    ),
    Panel.fit(job_progress, title="[b]Jobs", border_style="red", padding=(1, 2)),
)

with Live(progress_table, refresh_per_second=10):
    while not overall_progress.finished:
        sleep(0.1)
        for job in job_progress.tasks:
            if not job.finished:
                job_progress.advance(job.id)

        completed = sum(task.completed for task in job_progress.tasks)
        overall_progress.update(overall_task, completed=completed)



from pathlib import Path
from typing import Optional

import rich.progress
import typer
from click.exceptions import UsageError
from requests import HTTPError
from xplane_airports.AptDat import AptDat
from xplane_airports.gateway import scenery_pack

from .base import SUPPORTED_DRIVERS, VALID_FEATURES, ParsedAirport
from .geometry import _DEFAULT_BEZIER_RESOLUTION

app = typer.Typer(
    add_completion=False,
    pretty_exceptions_show_locals=False,
    pretty_exceptions_short=True,
)


@app.command()
def convert(
    airport_ids: str = typer.Option(
        ...,
        "--airport-ids",
        "-a",
        help="Comma separated list of airport IDs of the airports to parse and convert "
        + "(note that these are X-Plane airport IDs, usually matching the ICAO codes, but not always).",
    ),
    input_file: Optional[Path] = typer.Option(
        None,
        "--input-file",
        "-i",
        help="Input airport data (apt.dat) file path.",
        file_okay=True,
        dir_okay=False,
        readable=True,
        exists=True,
    ),
    download_recommended: bool = typer.Option(
        False,
        "--download-recommended",
        "-g",
        help="Download recommended airport files from the X-Plane Scenery Gateway "
        + "instead of using a local file.",
    ),
    output_folder: Path = typer.Option(
        "./",
        "--output-folder",
        "-o",
        help="Output folder.",
        file_okay=False,
        dir_okay=True,
        writable=True,
        exists=False,
    ),
    driver: str = typer.Option(
        "GeoJSON",
        "--driver",
        "-d",
        help=f"Format driver to use for the output. Supported drivers: {', '.join(SUPPORTED_DRIVERS.keys())}.",
        case_sensitive=False,
    ),
    crs: str = typer.Option(
        "EPSG:4326",
        "--crs",
        "-c",
        help="Coordinate reference system (CRS) to use for the output.",
    ),
    features: str = typer.Option(
        ",".join(VALID_FEATURES),
        "--features",
        "-f",
        help="Comma separated list of feature layers to include in the output.",
    ),
    bezier_resolution: int = typer.Option(
        _DEFAULT_BEZIER_RESOLUTION,
        "--bezier-resolution",
        "-z",
        help="Number of points to use to plot Bezier curves. "
        + "A higher number means more resolution but also larger file sizes.",
    ),
):
    """
    Convert X-Plane airport data to GIS-friendly formats like GeoJSON or ESRI Shapefile.

    Airports will either be loaded from a local apt.dat file (when using '--input-file' / '-i')
    or downloaded from the X-Plane Scenery Gateway (when using '--download-recommended' / '-g').
    """
    if airport_ids is None or airport_ids == "":
        raise UsageError("List of airport ids cannot be empty.")

    if input_file is not None and download_recommended:
        raise UsageError(
            "Only one of --input_file and --download_recommended can be used."
        )

    if input_file is None and not download_recommended:
        raise UsageError("Either --input_file or --download_recommended must be used.")

    apt_dat = None

    if input_file is not None:
        with rich.progress.open(input_file, "r") as f:
            apt_dat = AptDat.from_file_text(f.read(), input_file)

    for airport_id in rich.progress.track(
        airport_ids.split(","), description="Processing airports..."
    ):
        if apt_dat is not None:
            apt = apt_dat.search_by_id(airport_id)

            if apt is None:
                # logger.error(f"Airport {airport_id} not found in {input_file}.")
                print(f"Airport '{airport_id}' not found in {input_file}.")
                raise typer.Abort()

        else:
            assert download_recommended
            try:
                recommended_pack = scenery_pack(airport_id, retries_on_error=0)
            except HTTPError as e:
                print(e.__dict__)
                print(f"Failed to download '{airport_id}' ({e}).")
                continue

            apt = recommended_pack.apt

        extension = SUPPORTED_DRIVERS[driver]["default_extension"]
        output_file = output_folder / f"{airport_id}.{extension}"

        p_apt = ParsedAirport(apt, bezier_resolution=bezier_resolution)
        p_apt.export(output_file, driver=driver, crs=crs, features=features.split(","))
